import numpy as np
import soundfile as sf
from scipy.signal import stft
from scipy.signal import istft
import matplotlib.pyplot as plt

from calc_array_manifold_vector_solution import calcArrayManifoldVectorSln
from createIncomingWavFiles import createIncomingWavFiles

# Define parameters (you can play with these)
iElNum         =   8
fElSpacing     = 0.15
iIncomingDeg   = 30
iDesiredDirDeg = -80
iNoiseDirDeg   = -80

# Define other parameters
fSamplerate    = 11025
iNfft          = 512
iFrameshift    = 256
fSoundspeed    = 343
fTheta         = 180

# Frequency axis
pfFreqVec = np.linspace(0, fSamplerate/2, int(iNfft/2+1))

# Convert degree to rad
fTheta = np.arange(-round(fTheta/2),round(fTheta/2)+1)
fThetaRad = fTheta * np.pi / 180
iIncomingIndex = round(len(fTheta) / 2 + iIncomingDeg)
iDesiredDirIndex = round(len(fTheta) / 2 + iDesiredDirDeg)
iNoiseDirIndex = round(len(fTheta) / 2 + iNoiseDirDeg)

# Create a linear array (Slide 15.)
pfElPos = np.zeros((iElNum))
for iEl in np.arange(iElNum):
    pfElPos[iEl] = ( iEl - (iElNum - 1) / 2) * fElSpacing


# Create .wav files of incoming plane wave (transmit beamforming + noise)
createIncomingWavFiles(iFrameshift, iNfft, iElNum, iIncomingIndex, iNoiseDirIndex, pfElPos, pfFreqVec, fThetaRad, fSoundspeed)


# Calculate actual beampattern
pppcBeampatternRx, pppcArrayManifoldVectorRx = calcArrayManifoldVectorSln(iElNum, pfElPos, 'delsum', pfFreqVec, fThetaRad, fSoundspeed, 'rx')
pppcBeampatternBlock, pppcArrayManifoldVectorBlock = calcArrayManifoldVectorSln(iElNum, pfElPos, 'blocking', pfFreqVec, fThetaRad, fSoundspeed, 'rx')

# Convert complex beampattern for plotting purposes
pppfBeampatternRx20LogAbs = 20*np.log10(np.abs(pppcBeampatternRx + np.finfo(float).eps))
pppfBeampatternBlock20LogAbs = 20*np.log10(np.abs(pppcBeampatternBlock + np.finfo(float).eps))


# Read in audiofiles
iFileLengthInSamp = len(sf.SoundFile('bf_0.wav'))
pppfSignal = np.zeros((iElNum, iFileLengthInSamp))
ppcSignal = np.zeros((iElNum, int(iNfft/2+1), int(np.ceil(iFileLengthInSamp / iFrameshift ) + 1)), dtype=np.complex_)

# STFT of observed signal
for iEl in np.arange(iElNum):
    pppfSignal[iEl,:], fSamplerate = sf.read('bf_' + str(iEl) + '.wav')
    f_y, t_y, ppcSignal[iEl,:,:] = stft(pppfSignal[iEl,:], fs=fSamplerate, nperseg=iNfft, noverlap=iNfft-iFrameshift)

# Compute output spectrum
ppcSignalBf = np.zeros(np.shape(ppcSignal), dtype=np.complex_)
for iEl in np.arange(iElNum):
    for iFrame in np.arange(t_y.shape[0]):
        # Multiply each frame with the array manifold vector of the desired direction
        ppcSignalBf[iEl,:,iFrame] = np.multiply(pppcArrayManifoldVectorRx[iDesiredDirIndex,iEl,:],ppcSignal[iEl,:,iFrame])

# ISTFT of output signal
pfSignalBf = np.zeros((int(np.ceil(iFileLengthInSamp / iFrameshift )) * iFrameshift))
t_y, pfSignalBf = istft(np.sum(ppcSignalBf, axis=0), fs=fSamplerate, nperseg=iNfft, noverlap=iNfft-iFrameshift)

# Save .wav files
sf.write('bf_receive.wav', pfSignalBf, fSamplerate)

# Plot beampattern
plt.figure(42)
cmap = plt.get_cmap('jet')
plt.imshow(pppfBeampatternRx20LogAbs, aspect='auto', cmap=cmap, origin='lower', extent=[pfFreqVec[0], pfFreqVec[-1], fTheta[0], fTheta[-1]])
plt.clim(-40, 0)
plt.colorbar()

plt.figure(43)
cmap = plt.get_cmap('jet')
plt.imshow(pppfBeampatternBlock20LogAbs, aspect='auto', cmap=cmap, origin='lower', extent=[pfFreqVec[0], pfFreqVec[-1], fTheta[0], fTheta[-1]])
plt.clim(-40, 0)
plt.colorbar()
plt.show()